using System;
using System.Collections;
using System.Collections.Generic;
using System.Data;
using System.Reflection;

namespace FaultTolerantWebService.Common
{
    /// <summary>
    /// Classe utilitaire essayant de combler les petites lacunes des interfaces et classe
    /// de System.Collections.
    /// 
    /// Les cas les plus frquent sont li aux collections nulles. Toutes les methodes
    /// essaie de faire qu'un collection null est strictement equivalent  une collection
    /// vide.
    /// </summary>
    public static class CollectionUtil
    {
        /// <summary>
        /// Retourne le nombre d'lment, mme si l'argument est null :)
        /// 
        /// </summary>
        /// <param name="collection"></param>
        /// <typeparam name="ElementType"></typeparam>
        /// <returns></returns>
        public static int Count<ElementType>(ICollection<ElementType> collection)
        {
            if (collection != null) return collection.Count;
            return 0;
        }

        /// <summary>
        /// Fait la mme chose que string.IsNullOrEmpty mais sur des collections
        /// </summary>
        /// <typeparam name="ElementType"></typeparam>
        /// <param name="collection"></param>
        /// <returns></returns>
        public static bool IsNullOrEmptyG<ElementType>(ICollection<ElementType> collection)
        {
            return collection == null || collection.Count == 0;
        }

        /// <summary>
        /// Fait la mme chose que string.IsNullOrEmpty mais sur des collections
        /// </summary>
        /// <param name="collection"></param>
        /// <returns></returns>
        public static bool IsNullOrEmpty(ICollection collection)
        {
            return collection == null || collection.Count == 0;
        }

        /// <summary>
        /// Retourne le premier element d'une objet enumerable
        /// </summary>
        /// <typeparam name="ElementType"></typeparam>
        /// <param name="collection"></param>
        /// <returns></returns>
        public static ElementType GetFirst<ElementType>(IEnumerable<ElementType> collection)
        {
            if (collection != null) foreach (ElementType el in collection) return el;
            return default(ElementType);
        }

        /// <summary>
        /// Retourne N ime element d'un objet iterable
        /// (util pour dbugger des IDictionary)
        /// </summary>
        /// <typeparam name="ElementType"></typeparam>
        /// <param name="collection"></param>
        /// <param name="position"></param>
        /// <returns></returns>
        public static ElementType GetNth<ElementType>(IEnumerable<ElementType> collection, int position)
        {
            if (collection == null) return default(ElementType);
            if (position < 0) return default(ElementType);

            int i = 0;
            foreach (ElementType el in collection)
            {
                if (i == position) return el;
                i++;
            }

            return default(ElementType);
        }

        /// <summary>
        /// Retourne un tableau  partir d'un object enumrable (ce qui inclue les objets implmentant IEnumerable, ICollection, IDictionnary, IList, ISet ... et les tableaux)
        /// </summary>
        /// <typeparam name="ElementType"></typeparam>
        /// <param name="iList"></param>
        /// <returns></returns>
        public static ElementType[] ToArray<ElementType>(IEnumerable<ElementType> elements)
        {
            return new List<ElementType>(elements ?? new ElementType[] { }).ToArray();
        }

        /// <summary>
        /// Retourne une Liste gnrique  partir d'un objet numrable 
        /// (ce qui inclue les objets implmentant IEnumerable, ICollection, IDictionnary, IList, ISet ... et les tableaux)
        /// </summary>
        /// <typeparam name="ElementType"></typeparam>
        /// <param name="elements"></param>
        /// <returns></returns>
        public static List<ElementType> ToList<ElementType>(IEnumerable<ElementType> elements)
        {
            return new List<ElementType>(elements ?? new ElementType[] { });
        }

        /// <summary>
        /// Determine si l'aiguile (needle) se trouve dans la bote de foins :)
        /// </summary>
        /// <param name="haystack">la boite de foins</param>
        /// <param name="needle">L'aiguille que l'on recherche</param>
        /// <returns></returns>
        public static bool Contains(IEnumerable haystack, object needle)
        {
            foreach (object currentElement in haystack ?? new object[] { })
            {
                if (Equals(needle, currentElement))
                    return true;
            }
            return false;
        }



        /// <summary>
        /// Retourne une liste trie avec la methode de trie correspondant
        /// </summary>
        /// <typeparam name="TElement"></typeparam>
        /// <param name="elements"></param>
        /// <param name="comparisonMethod"></param>
        /// <returns></returns>
        public static List<TElement> ToSortedList<TElement>(ICollection<TElement> elements, Comparison<TElement> comparisonMethod)
        {
            if (elements == null) return null;
            TElement[] buffer = GetArrayBuffer<TElement>(elements);
            Array.Sort<TElement>(buffer, comparisonMethod);
            return new List<TElement>(buffer);
        }

        /// <summary>
        /// Retourne une liste trie avec la methode de trie correspondant
        /// </summary>
        /// <typeparam name="TElement"></typeparam>
        /// <param name="elements"></param>
        /// <param name="comparisonMethod"></param>
        /// <returns></returns>
        public static List<TElement> ToSortedList<TElement>(ICollection<TElement> elements, IComparer<TElement> comparer)
        {
            if (elements == null) return null;
            TElement[] buffer = GetArrayBuffer<TElement>(elements);
            Array.Sort<TElement>(buffer, comparer);
            return new List<TElement>(buffer);
        }

        /// <summary>
        /// Retourne une liste trie moyennat que les elements soient comparables
        /// </summary>
        /// <typeparam name="TElement"></typeparam>
        /// <param name="elements"></param>
        /// <param name="comparisonMethod"></param>
        /// <returns></returns>
        public static List<TElement> ToSortedList<TElement>(ICollection<TElement> elements) where TElement : IComparable
        {
            if (elements == null) return null;

            TElement[] buffer = GetArrayBuffer<TElement>(elements);
            Array.Sort<TElement>(buffer);
            return new List<TElement>(buffer);
        }


        private static TElement[] GetArrayBuffer<TElement>(ICollection<TElement> elements)
        {
            if (elements == null) elements = new List<TElement>();

            TElement[] buffer = new TElement[elements.Count];
            int i = 0; foreach (TElement el in elements) buffer[i++] = el;
            return buffer;
        }

        /// <summary>
        /// Retourne le ToString() de chaque lment spar par une virgule
        /// </summary>
        /// <param name="collection"></param>
        /// <returns></returns>
        public static string ToString(IEnumerable collection)
        {
            if (collection == null) return "";
            List<string> result = new List<string>();
            foreach (object el in collection)
                result.Add((el ?? "").ToString());
            return string.Join(", ", result.ToArray());
        }
    }
}