using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.Net;
using log4net;
using System.Threading;
using FaultTolerantWebService.Common.Configuration;
using System.Configuration;
using System.Web.Services.Protocols;

namespace FaultTolerantWebService.Common
{
    /// <summary>
    /// Classe qui permet d'invoker les methodes de la mre ayant le mme nom
    /// La methode  apelle est <see cref="TryInvoke"/>
    /// </summary>
    public class Invoker
    {
        #region singleton

        private static Invoker instance;
        /// <summary>
        /// Instance par dfaut avec des valeurs de config par dfaut
        /// </summary>
        public static Invoker Instance
        {
            get
            {
                if (instance == null) instance = new Invoker();
                return Invoker.instance;
            }
            set { Invoker.instance = value; }
        }

        #endregion

        private ILog logger = LogManager.GetLogger(typeof(Invoker));
        private ConfigSection configSection = null;
        private Dictionary<Type, List<ClassUri>> altUriCache = new Dictionary<Type, List<ClassUri>>();

        /// <summary>
        /// Nb maximum de tentative avant de laisser sortirt une exception
        /// </summary>
        public int MaxRetry
        {
            get { return GetConfigSection() != null ? GetConfigSection().MaxRetry : 5; }
        }

        /// <summary>
        /// Temps d'attente entre les acppels
        /// </summary>
        public int RetryInterval
        {
            get { return GetConfigSection() != null ? GetConfigSection().RetryInterval : 2000; }
        }

        public Invoker()
        {
        }

        /// <summary>
        /// Essaie de lancer une methode n fois. Si au bout de maxTryCount
        /// on contienue d'avoir une exception, alors, on laisse la dernire
        /// exception sortir.
        /// </summary>
        /// <param name="method"></param>
        /// <param name="args"></param>
        /// <returns></returns>
        public object TryInvoke(MethodInfo method, object target, object[] args)
        {
            Exception lastException = null;
            for (int i = 0; i < MaxRetry; i++)
            {
                try
                {
                    logger.Info(method.Name + " - i=" + i + ", args=" + TextUtil.Left(CollectionUtil.ToString(args), 2000));
                    return method.Invoke(target, args);
                }
                catch (Exception e)
                {
                    lastException = e;

                    //on essaie les autres url
                    IList<string> altUriList = GetAlternativeUriList((SoapHttpClientProtocol)target);
                    if (altUriList.Count > 0)
                        ((SoapHttpClientProtocol)target).Url = altUriList[i % altUriList.Count];

                    //on attend RetryInterval
                    logger.Warn(method.Name + " failed - i=" + i + ", args=" + TextUtil.Left(CollectionUtil.ToString(args), 2000));
                    Thread.Sleep(RetryInterval);
                }
            }
            throw (lastException);
        }

        /// <summary>
        /// Retourne la liste des uri alternatives pour un client soap
        /// </summary>
        /// <param name="type">Le type </param>
        /// <returns></returns>
        protected virtual IList<string> GetAlternativeUriList(SoapHttpClientProtocol soapClient)
        {
            List<string> result = new List<string>();
            ConfigSection section = GetConfigSection();
            if (section != null)
            {
                //on gnre le cache des uri
                if (altUriCache.Count == 0)
                {
                    //on prend toutes les uri et on indexe par soapClientClass
                    foreach (ClassUri cu in section.AlternativeUriList)
                    {
                        Type type = Type.GetType(cu.SoapClientClass);
                        if (!altUriCache.ContainsKey(type)) altUriCache[type] = new List<ClassUri>();
                        altUriCache[type].Add(cu);
                    }
                }

                //on prends mes Uri et ceux de mes ancetres
                List<ClassUri> mergedResult = new List<ClassUri>();
                Type currentType = soapClient.GetType();
                while (currentType != null)
                {
                    mergedResult.AddRange(GetStrictUri(currentType));
                    currentType = currentType.BaseType;
                }

                //on trie
                mergedResult.Sort(
                    delegate(ClassUri a, ClassUri b)
                    {
                        return b.Priority - a.Priority;
                    }
                );

                //on ne prend plus que les uri
                foreach (ClassUri item in mergedResult) result.Add(item.Uri);

                //pour simplifier ailleurs, on ne retourne jamais une liste vide
                if (result.Count == 0) result.Add(soapClient.Url);
            }
            return result;
        }

        private List<ClassUri> GetStrictUri(Type type)
        {
            List<ClassUri> result = new List<ClassUri>();
            if (!altUriCache.ContainsKey(type)) altUriCache[type] = new List<ClassUri>();
            foreach (ClassUri item in altUriCache[type])
                result.Add(item);
            return result;
        }

        protected virtual ConfigSection GetConfigSection()
        {
            if (configSection == null) configSection = (ConfigSection)ConfigurationManager.GetSection(GetType().FullName);
            return configSection;
        }
    }
}
