using System;
using System.Collections.Generic;
using System.Text;
using System.Reflection;
using System.CodeDom;
using System.IO;
using System.CodeDom.Compiler;
using System.Web.Services.Protocols;
using FaultTolerantWebService.Common;

namespace FaultTolerantWebService.Generator
{
    /// <summary>
    /// Point d'entre pour la gnration des proxy soap
    /// Il faut appeler la methode <see cref="GenerateProxy"/>
    /// </summary>
    public class Engine
    {
        private string language;
        /// <summary>
        /// Obtien ou dfinit le language de programmation .NET  utiliser pour la
        /// gnration (Par dfaut ce sera C#)
        /// </summary>
        public string Language
        {
            get
            {
                if (string.IsNullOrEmpty(language)) language = CodeDomProvider.GetLanguageFromExtension(".cs");
                return language;
            }
            set { language = value; }
        }

        public Engine()
        {
        }

        public Engine(string language)
        {
            this.language = language;
        }

        /// <summary>
        /// Retourne la liste des classes proxy de l'<see cref="assembly"/> 
        /// </summary>
        /// <param name="assembly"></param>
        /// <returns></returns>
        public virtual List<Type> GetSoapProxyList(Assembly assembly)
        {
            List<Type> result = new List<Type>();
            foreach (Type t in assembly.GetTypes())
            {
                if (t.IsClass && t.BaseType == typeof(SoapHttpClientProtocol))
                {
                    result.Add(t);
                }
            }
            return result;
        }

        /// <summary>
        /// Gnre un proxy "fault tolerant" pour chaque classe proxy soap de l'assembly <see cref="assembly"/>
        /// </summary>
        /// <param name="t"></param>
        /// <param name="output"></param>
        /// <returns></returns>
        public virtual Engine GenerateProxy(Assembly assembly, TextWriter output)
        {
            foreach (Type t in GetSoapProxyList(assembly))
                GenerateProxy(t, output);

            return this;
        }

        /// <summary>
        /// Gnre un proxy "fault tolerant" pour la classe proxy <see cref="t"/>
        /// </summary>
        /// <param name="t"></param>
        /// <param name="output"></param>
        /// <returns></returns>
        public virtual Engine GenerateProxy(Type t, TextWriter output)
        {
            Serialize(
                GenerateClassCode(t, GetStdMethods(t)),
                Language,
                output
            );
            return this;
        }

        protected virtual Engine Serialize(CodeCompileUnit codeCompileUnit, string language, TextWriter output)
        {
            CodeDomProvider compiler = CodeDomProvider.CreateProvider(language);
            CodeGeneratorOptions options = new CodeGeneratorOptions();
            options.BlankLinesBetweenMembers = false;
            compiler.GenerateCodeFromCompileUnit(
                codeCompileUnit,
                output,
                options
            );
            return this;
        }

        /// <summary>
        /// Gnre le code sous forme d'arborescence CodeDom
        /// 
        /// Voici grossomodo ce qu'on veut :
        /// namespace Oddo.Diamond.Crm.Core.FaultTolerantServices
        /// {
        ///     public class FaultTolerantReportService : ReportsManagement
        ///     {
        ///         public new byte[] GenerateCAR(string languageCode, int companyCode, int customerCode, System.DateTime startDate, DateTime endDate, int teamCode, bool equities, bool fixedIncome)
        ///         {
        ///             return (byte[]) Invoker.Instance.TryInvoke(
        ///                 GetType().BaseType.GetMethod(MethodInfo.GetCurrentMethod().Name),
        ///                 this,
        ///                 new object[] { languageCode, companyCode, customerCode, startDate, endDate, teamCode, equities, fixedIncome }
        ///             );
        ///         }
        ///     }
        /// }
        /// </summary>
        /// <param name="m"></param>
        /// <returns></returns>
        protected virtual CodeCompileUnit GenerateClassCode(Type t, List<MethodInfo> list)
        {
            CodeCompileUnit compileUnit = new CodeCompileUnit();
            CodeNamespace ns = new CodeNamespace(t.Namespace);
            compileUnit.Namespaces.Add(ns);

            CodeTypeDeclaration newClass = new CodeTypeDeclaration("FaultTolerant" + t.Name);
            newClass.BaseTypes.Add(new CodeTypeReference(t));
            newClass.Attributes = MemberAttributes.Public;
            newClass.Comments.Add(new CodeCommentStatement("<summary>", true));
            newClass.Comments.Add(new CodeCommentStatement("Fault tolerant soap proxy encapsulating a " + t.FullName, true));
            newClass.Comments.Add(new CodeCommentStatement("Generated by " + Assembly.GetExecutingAssembly().GetName().Name + " version " + Assembly.GetExecutingAssembly().GetName().Version, true));
            newClass.Comments.Add(new CodeCommentStatement("For news and updates goto : " + Constant.ProjectHomeUrl, true));
            newClass.Comments.Add(new CodeCommentStatement("</summary>", true));
            ns.Types.Add(newClass);

            foreach (MethodInfo method in list)
                newClass.Members.Add(GenerateMethodCode(method));

            return compileUnit;
        }

        /// <summary>
        /// Gnre le code d'une methode pour voir l'exemple voir <see cref="GenerateClassCode"/>
        /// </summary>
        /// <param name="m"></param>
        /// <returns></returns>
        protected virtual CodeMemberMethod GenerateMethodCode(MethodInfo m)
        {
            List<string> paramNameList = new List<string>();
            List<Type> paramTypeList = new List<Type>();
            List<CodeExpression> paramExpressionList = new List<CodeExpression>();

            CodeMemberMethod method = new CodeMemberMethod();

            //Type retour + modifier + paramettres
            method.Name = m.Name;
            method.ReturnType = new CodeTypeReference(m.ReturnType);
            method.Attributes = MemberAttributes.Public | MemberAttributes.Final | MemberAttributes.New;
            foreach (ParameterInfo param in m.GetParameters())
            {
                method.Parameters.Add(new CodeParameterDeclarationExpression(
                    new CodeTypeReference(param.ParameterType),
                    param.Name
                ));
                paramNameList.Add(param.Name);
                paramTypeList.Add(param.ParameterType);
                paramExpressionList.Add(new CodeVariableReferenceExpression(param.Name));
            }

            //statement : Invoker.Instance.TryInvoke(MethodInfo.GetCurrentMethod().Name, this,  ...
            CodeExpression maintStatement =
                new CodeMethodInvokeExpression(
                    new CodeFieldReferenceExpression(
                        new CodeTypeReferenceExpression(typeof(Invoker)),
                        "Instance"
                    ),
                    "TryInvoke",
                    new CodeExpression[]{
                        new CodeMethodInvokeExpression(
                            new CodeFieldReferenceExpression(
                                new CodeMethodInvokeExpression(
                                    new CodeMethodReferenceExpression(
                                        new CodeThisReferenceExpression(),
                                        "GetType"
                                    )
                                ),
                                "BaseType"
                            ),                                
                            "GetMethod",
                            new CodeExpression[] {
                                new CodeFieldReferenceExpression(
                                    new CodeMethodInvokeExpression(
                                        new CodeTypeReferenceExpression(typeof(MethodInfo)),
                                        "GetCurrentMethod"
                                    ),
                                    "Name"
                                )
                            }
                        ),
                        new CodeThisReferenceExpression(),
                        new CodeArrayCreateExpression(
                            typeof(object),
                            paramExpressionList.ToArray()
                        )
                    }
                );

            //Selon que ce soit void ou non, on fait un return + cast
            if (m.ReturnType != typeof(void))
            {
                method.Statements.Add(
                    new CodeMethodReturnStatement(
                        new CodeCastExpression(
                            new CodeTypeReference(m.ReturnType),
                            maintStatement
                        )
                    )
                );
            }
            else
            {
                method.Statements.Add(new CodeExpressionStatement(maintStatement));
            }

            return method;
        }

        /// <summary>
        /// Retourne la liste des mthodes tant des appel synchrones vers un webservice
        /// </summary>
        /// <param name="t"></param>
        /// <returns></returns>
        protected virtual List<MethodInfo> GetStdMethods(Type t)
        {
            List<MethodInfo> methods = new List<MethodInfo>();
            foreach (MethodInfo m in t.GetMethods())
            {
                object[] attributes = m.GetCustomAttributes(typeof(SoapDocumentMethodAttribute), false);
                if (attributes.Length > 0) methods.Add(m);
            }
            return methods;
        }
    }
}
